from .alerts import AlertsService
from .base import BaseService
from .dashboards import DashboardsService
from .data_sources import DataSourcesService
from .destinations import DestService
from .groups import GroupsService
from .mixins import *
from .queries import QueriesService
from .query_snippets import QSnipsService
from .users import UsersService



from typing import Dict, List, Optional

from .base import BaseService
from .mixins import (
    CommonMixin,
    FavoriteMixin,
    NameMixin,
    PaginationMixin,
    PrintMixin,
    PublishMxin,
    TagsMixin,
)


class DashboardsService(
    FavoriteMixin,
    CommonMixin,
    TagsMixin,
    PublishMxin,
    NameMixin,
    PrintMixin,
    PaginationMixin,
):
    def __init__(self, base: BaseService) -> None:

        # init mixins
        FavoriteMixin.__init__(self, base)
        CommonMixin.__init__(self, base)
        PublishMxin.__init__(self, base)
        PaginationMixin.__init__(self, base)

        self.__base = base
        self.endpoint = "/api/dashboards"

    def get_all(self) -> List[Dict]:
        return self.paginate()

    def get_slug(self, dashboard_id: int) -> Optional[str]:
        """Get the slug for a dashboard by ID"""
        return self.get(dashboard_id).get("slug")

    def refresh(self, dashboard_id: int) -> None:
        """Refresh a dashboard"""
        widgets = self.get(dashboard_id).get("widgets")

        for widget in widgets:
            if not "visualization" in widget.keys():
                continue
            query = widget.get("visualization").get("query")
            self.__base.post(f"/api/queries/{query.get('id')}/results", {"max_age": 0})

    def share(self, dashboard_id: int) -> str:
        """get public url for dashboard"""
        response = self.__base.post(f"{self.endpoint}/{dashboard_id}/share", {})
        return response.get("public_url")

    def duplicate(self, dashboard_id: int, new_name: Optional[str] = None) -> Dict:
        """Duplicate a dashboard and all its widgets with `new_name`"""
        current = self.get(dashboard_id)

        if new_name is None:
            new_name = f"Copy of: {current.get('name')}"

        new_dash = self.create({"name": new_name})

        if current.get("tags") is not None:
            self.update(new_dash.get("id"), {"tags": current.get("tags")})

        for widget in current.get("widgets"):
            visualization_id = None
            if "visualization" in widget.keys():
                visualization_id = widget.get("visualization").get("id")

            self.create_widget(
                dashboard_id=new_dash.get("id"),
                visualization_id=visualization_id,
                options=widget.get("options"),
                text=widget.get("text"),
            )

    def create_widget(
        self,
        *,
        dashboard_id: int,
        visualization_id: Optional[int],
        options: Dict,
        text: str = "",
    ) -> Dict:
        """
        create new widget in given dashboard

        Args:
            dashboard_id: id of dashboard to create widget in
            visualization_id: id of visualization to use for widget (pass None for text widget)
            options: options to use for widget
            text: text to use for text widget
        """
        data = dict(
            dashboard_id=dashboard_id,
            text=text,
            options=options,
            visualization_id=visualization_id,
            width=1,
        )
        return self.__base.post("/api/widgets", data)



from typing import Dict, Optional, final

import requests


class BaseService:
    """Singleton Base Service class"""

    def __init__(self, base_url: str, api_key: str) -> None:
        self.base_url = base_url

        # configure session
        self.__session = requests.Session()
        self.__session.headers.update({"Authorization": f"Key {api_key}"})
        self.__session.headers.update({"Content-Type": "application/json"})
        self.__session.headers.update({"Accept": "application/json"})

    @final
    def _request(
        self,
        method: str,
        endpoint: str,
        data: Optional[Dict] = None,
        **kwargs,
    ) -> requests.Response:
        """Perform a request to the endpoint with the given data"""
        url = self.base_url + endpoint
        response = self.__session.request(
            method,
            url,
            json=data if data else None,
            **kwargs,
        )
        response.raise_for_status()
        return response

    @final
    def post(self, endpoint: str, data: Dict, **kwargs) -> Dict:
        """Perform a POST request to the endpoint with the given data"""
        return self._request("POST", endpoint, data, **kwargs).json()

    @final
    def get(self, endpoint: str, data: Optional[Dict] = None, **kwargs) -> Dict:
        """Perform a GET request to the endpoint"""
        return self._request("GET", endpoint, data, **kwargs).json()

    @final
    def delete(self, endpoint: str, **kwargs) -> None:
        """Perform a DELETE request to the endpoint"""
        return self._request("DELETE", endpoint, **kwargs)



from typing import Dict, List, Optional, final

from .base import BaseService


class PrintMixin:
    """Mixin class for printing data"""

    @final
    def __repr__(self) -> str:
        object_methods = [
            method_name
            for method_name in dir(self)
            if callable(getattr(self, method_name)) and not method_name.startswith("_")
        ]
        object_attributes = [
            attribute_name
            for attribute_name in dir(self)
            if not callable(getattr(self, attribute_name))
            and not attribute_name.startswith("_")
        ]
        return f"{self.__class__.__name__}(attributes: {object_attributes}, methods: {object_methods})"

    @final
    def __str__(self) -> str:
        return self.__repr__()


class CommonMixin:
    """Mixin with common methods for services."""

    def __init__(self, base: BaseService) -> None:
        self.__base = base

    def get(self, id: int) -> Dict:
        """Fetch one by ID"""
        return self.__base.get(f"{self.endpoint}/{id}")

    def get_all(self) -> List[Dict]:
        """fetch all objects."""
        return self.__base.get(self.endpoint)

    def update(self, id: int, data: Dict) -> Dict:
        """Update by ID"""
        return self.__base.post(f"{self.endpoint}/{id}", data)

    def create(self, data: Dict) -> Dict:
        """Create a new object with data"""
        return self.__base.post(self.endpoint, data)

    def delete(self, id: int) -> None:
        """Delete by ID"""
        return self.__base.delete(f"{self.endpoint}/{id}")


class NameMixin:
    """Mixin for nameable services (by name or slug)."""

    def exists(self, name_or_slug: str) -> bool:
        """Check if an object with given `name_or_slug` exists"""
        return self.get_id(name_or_slug) is not None

    def get_by_name(self, name_or_slug: str) -> Dict:
        """Get by name or slug"""
        return self.get(self.get_id(name_or_slug))

    def get_id(self, name_or_slug: str) -> Optional[int]:
        """Get the ID for an object by name or slug, returns None if not found"""
        all_obj: List[Dict] = self.get_all()

        if "results" in all_obj:
            all_obj = all_obj["results"]

        for obj in all_obj:
            if obj.get("name") == name_or_slug or obj.get("slug") == name_or_slug:
                return obj.get("id")

        return None


class TagsMixin:
    """Mixin for taggable services"""

    def get_by_tags(self, tags: List[str], without: bool = False) -> List[Dict]:
        """Get all objects with `tags` or all objects without any of `tags`"""
        all_objects = self.get_all()

        if "results" in all_objects:
            all_objects = all_objects["results"]

        if without:
            return [
                obj
                for obj in all_objects
                if not any(tag in obj.get("tags", []) for tag in tags)
            ]

        return [
            obj
            for obj in all_objects
            if any(tag in obj.get("tags", []) for tag in tags)
        ]


class PublishMxin:
    """Mixin for publishable services"""

    def __init__(self, base: BaseService) -> None:
        self.__base = base

    def publish(self, dashboard_id: int) -> Dict:
        """Publish an object"""
        return self.__base.post(f"{self.endpoint}/{dashboard_id}", {"is_draft": False})

    def unpublish(self, dashboard_id: int) -> Dict:
        """Unpublish an object"""
        return self.__base.post(f"{self.endpoint}/{dashboard_id}", {"is_draft": True})


class FavoriteMixin:
    """Mixin for favoriteable services"""

    def __init__(self, base: BaseService) -> None:
        self.__base = base

    def favorited(self) -> List[Dict]:
        """Get all favorited objects"""
        return list(filter(lambda x: x["is_favorite"], self.get_all()))

    def favorite(self, id: int) -> Dict:
        """Favorite an object"""
        return self.__base.post(f"{self.endpoint}/{id}/favorite", {})

    def unfavorite(self, id: int) -> Dict:
        """Unfavorite an object"""
        return self.__base.delete(f"{self.endpoint}/{id}/favorite")


class PaginationMixin:
    """Mixin for paginated services"""

    def __init__(self, base: BaseService) -> None:
        self.__base = base

    def paginate(self, page: int = 1, page_size: int = 100, **kwargs) -> List[Dict]:
        """Load all items of a paginated resource"""
        response = self.__base.get(
            self.endpoint, {"page": page, "page_size": page_size}, **kwargs
        )
        items = response.get("results")
        if response.get("page") * response.get("page_size") >= response.get("count"):
            return items

        else:
            return [
                *items,
                *self.paginate(page + 1, page_size, **kwargs),
            ]



from .base import BaseService
from .mixins import CommonMixin, NameMixin, PrintMixin


class GroupsService(CommonMixin, NameMixin, PrintMixin):
    def __init__(self, base: BaseService) -> None:

        # init mixins
        CommonMixin.__init__(self, base)

        self.endpoint = "/api/groups"



from typing import Dict, List, Optional

from .base import BaseService
from .mixins import (
    CommonMixin,
    FavoriteMixin,
    NameMixin,
    PaginationMixin,
    PrintMixin,
    PublishMxin,
    TagsMixin,
)


class QueriesService(
    FavoriteMixin,
    CommonMixin,
    TagsMixin,
    PublishMxin,
    NameMixin,
    PrintMixin,
    PaginationMixin,
):
    def __init__(self, base: BaseService) -> None:

        # init mixins
        FavoriteMixin.__init__(self, base)
        CommonMixin.__init__(self, base)
        PublishMxin.__init__(self, base)
        PaginationMixin.__init__(self, base)

        self.__base = base
        self.endpoint = "/api/queries"

    def get_all(self) -> List[Dict]:
        return self.paginate()

    def refresh(self, query_id: int) -> Dict:
        """Refresh a query"""
        return self.__base.post(f"{self.endpoint}/{query_id}/results", dict(max_age=0))

    def fork(self, query_id: int, new_name: str = None) -> Dict:
        """Fork a query with a new name if provided"""
        new_q = self.__base.post(f"{self.endpoint}/{query_id}/fork", {"id": query_id})

        if not new_name:
            return new_q

        new_q["name"] = new_name
        return self.update(new_q["id"], new_q)

    def scheduled(self) -> List[Dict]:
        """Get all scheduled queries"""
        return list(filter(lambda q: q.get("schedule"), self.get_all()))

    def update_visualization(self, vis_id: int, data: Dict) -> Dict:
        """Update a visualization with `data`"""
        return self.__base.post(f"/api/visualizations/{vis_id}", data)

    def duplicate_query_table(
        self,
        *,
        query: Dict,
        new_name: str = None,
        table_map: Dict[str, str],
        tags: List[str],
        publish: bool = True,
    ) -> Dict:
        """
        Duplicate query with source tables replaced according to table_map.

        Args:
            query(Dict): query data object
            new_name(str): new query name (optional)
            table_map(Dict): mapping of old to new tables
            tags(List): tags to add to new query
            publish(bool): whether to publish the new query

        Returns:
            Dict: new query data object

        Example:
            >>> query = queries.get(query_id)
            >>> table_map = {
            >>>     "old_table": "new_table",
            >>>     "old_table2": "new_table2",
            >>> }
            >>> tags = ["tag1", "tag2"]
            >>> new_query = queries.duplicate_query_table(
            >>>     query=query, table_map=table_map, tags=tags, publish=True, new_name="new_query_name"
            >>> )
        """

        new_id = self.fork(query.get("id"), new_name).get("id")

        # update query
        new_query = self.get(new_id)
        new_query["name"] = query.get("name")
        new_query["tags"] = tags

        for old, new in list(table_map.items()):
            new_query["query"] = new_query.get("query").replace(old, new)
        self.update(new_id, new_query)

        # run updated query to update results
        self.refresh(new_query.get("id"))

        # finally publish the new query
        if publish:
            self.publish(new_id)

        return self.get(new_id)



from .base import BaseService
from .mixins import CommonMixin, NameMixin, PrintMixin


class UsersService(CommonMixin, NameMixin, PrintMixin):
    def __init__(self, base: BaseService) -> None:

        # init mixins
        CommonMixin.__init__(self, base)

        self.__base = base
        self.endpoint = "/api/users"

    def disable(self, user_id: int) -> None:
        """
        disable user
        """
        self.__base.post(f"/api/users/{user_id}/disable")

    def add_group(self, user_id: int, group_id: int) -> None:
        """
        add user to group
        """
        self.__base.post(f"/api/groups/{group_id}/members", {"user_id": user_id})

    def remove_group(self, user_id: int, group_id: int) -> None:
        """
        remove user from group
        """
        self.__base.delete(f"/api/groups/{group_id}/members/{user_id}")



from typing import List

import requests

from redash_python.services import (
    AlertsService,
    BaseService,
    DashboardsService,
    DataSourcesService,
    DestService,
    GroupsService,
    QSnipsService,
    QueriesService,
    UsersService,
)

version = "0.2.21"


class Redash:
    """
    Top Class of The Redash API Client hierarchy.

    Args:
        base_url (str): The base URL of the Redash instance.
        api_key (str): The API key for the Redash instance.

    Usage:
        >>> from redash_python import Redash
        >>> rd = Redash("https://redash.example.com", "API_KEY")
        >>> rd.dashboards.get_all()
    """

    def __init__(self, base_url: str, api_key: str) -> None:
        self.version = version

        self.__base = BaseService(base_url, api_key)

        # Initialize services
        self.dashboards = DashboardsService(self.__base)
        self.queries = QueriesService(self.__base)
        self.data_sources = DataSourcesService(self.__base)
        self.alerts = AlertsService(self.__base)
        self.destinations = DestService(self.__base)
        self.users = UsersService(self.__base)
        self.groups = GroupsService(self.__base)
        self.query_snippets = QSnipsService(self.__base)

        # keep track of services
        attr_dict = self.__dict__
        ignore_list = ["_Redash__base", "__base", "version"]
        self.services = [
            attr_dict[k].__class__.__name__
            for k in self.__dict__
            if k not in ignore_list
        ]
        self.services.sort()

        # make __base methods available
        self.get = self.__base.get
        self.post = self.__base.post
        self.delete = self.__base.delete

    def test_credentials(self):
        """Test whether current credentials are valid"""
        try:
            response = self.get("/api/session")
            return True
        except requests.exceptions.HTTPError:
            return False

    def __dir__(self) -> List[str]:
        return self.services + ["get", "post", "delete"]

    def __repr__(self) -> str:
        return f"Redash(base_url: {self.__base.base_url}, services: {self.services})"

    def __str__(self) -> str:
        return self.__repr__()
